#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define NMAX 100000               /* maximal number of bodies present */
#define SIZE 100                       /* maximal length of file name */
#define NTAB 32

struct solar                                     /* single body data: */
   {
      double m;                                               /* mass */
      double r;                                             /* radius */
      double rs;                                      /* sqrt(radius) */
      double ris;                                 /* 1 / sqrt(radius) */
      double s;                                               /* spin */
      double rdown;                       /* left attraction boundary */
      double rup;                        /* right attraction boundary */
   };

struct solar v[NMAX];
long idum2 = 123456789;                            /* random number   */
long iy = 0, iv[NTAB];                             /* generator stuff */

void solar(long iN, long iK, long irep, double N, double K,
           long *seed, int save, double Nsave, int snap,
           int mode_snap, double Nsnapmin, double Nsnapstep,
           long Nsnaprep, char *output_name)

                  /* accretes one system, no avelanches, left merging */
                                                          /* ver. 4.1 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */

   {
      double rho(long*);
      int cond_dust(long, struct solar, long*, double);
      void shift_add(long, struct solar, long*);
      void sort2(unsigned long, double*, long*);
      int save_or_not(long*, double);

      void save_crash_file(long, long, long, double, double, long,
                           double, long, double, long, long);

      int snap_or_not(long*, double, int, double, double, long);
      void save_snap_file(double, double, long, double, long);

      struct solar dust;                        /* dust particle data */
      long isave = 0, tsnap = 0, j;                       /* counters */
      double i; 
      long n = 0;                         /* current number of bodies */
      long down, center, up;                         /* sorting stuff */
      long index_arr[NMAX];
      double sort_arr[NMAX];
      FILE *output_file;

      if(output_name[0] == '\n')
         {
            output_file = stdout;
         }

      else
         {
            output_file = fopen(output_name, "a+");
         }

      dust.m = 1 / N;                       /* initialization of dust */
      dust.r = rho(seed);
      dust.rs = sqrt(dust.r);
      dust.ris = 1 / dust.rs;
      dust.s = 0;
      dust.rdown = dust.r;
      dust.rup = dust.r;

      v[0] = dust;                                  /* the first body */

                                             /* main loop over N - 1  */
      for(i = 1; i < N; ++i)                 /* dust particles to add */
         {
            if(save == 1)      /* saving to a crash file ? */
               {
                  if(save_or_not(&isave, Nsave) == 1)
                     {
                        save_crash_file(iN, iK, irep, N, K, n, Nsave,
                                        *seed, i, isave - 1, tsnap);
                     }
               }

            if(snap == 1)                      /* making a snapshot ? */
               {
                  if(snap_or_not(&tsnap, i, mode_snap,
                                 Nsnapmin, Nsnapstep, Nsnaprep) == 1)
                     {
                        save_snap_file(N, K, n, i, irep);
                     }
               }

            dust.r = rho(seed);                 /* next dust particle */
            dust.rs = sqrt(dust.r);
            dust.ris = 1 / dust.rs;
            dust.rdown = dust.r;
            dust.rup = dust.r;

            if(dust.r <= v[0].r)      /* left from the left most body */
               {
                  if(cond_dust(0, dust, &n, K))  /* merging ?         */
                     {                           /* if yes, check     */
                        continue;                /* for further       */
                     }                           /* possibilities     */

                  else
                     {
                        shift_add(0, dust, &n);  /* if not, just add  */
                        continue;                /* particle          */
                     }
               }

            if(dust.r >= v[n].r)   /* right from the right most body */
               {
                  if(cond_dust(n, dust, &n, K))    /* merging ?       */
                     {                             /* if yes, check   */
                        continue;                  /* for further     */
                     }                             /* possibilities   */

                  else
                     {
                        shift_add(n + 1, dust, &n);  /* if not, just  */
                        continue;                    /* add particle  */
                     }
               }

            up = n;       /* locate the position of dust by bisection */
            down = 0;

            while(up - down > 1)
               {
                  center = (up + down) / 2;
                  
                  if(dust.r < v[center].r)
                     {
                        up = center;
                     }
                     
                  else
                     {
                        down = center;
                     }                     /* dust is between v[down] */
               }                           /* and v[up = down + 1]    */

            if(cond_dust(down, dust, &n, K)) /* merging to the left ? */
               {                              /* if yes, check        */
                  continue;                   /* for further          */
               }                              /* possibilities        */

            if(cond_dust(down + 1, dust, &n, K)) /* merging to the    */
               {                                 /* right ? if yes,   */
                  continue;                      /* check for further */
               }                                 /* possibilities     */

            shift_add(down + 1, dust, &n);     /* if neither is true, */
                                               /* just add particle   */

         }                                    /* end of the main loop */

      for(j = 0; j < n + 1; ++j)
         {
            index_arr[j + 1] = j;
            sort_arr[j + 1] = v[j].m;
         }

      sort2(n + 1, sort_arr, index_arr);      /* sorting in ascending */
                                              /* mass for output      */

      fprintf(output_file, "\nN=%le\nK=%le\nn=%ld number of final particles\nradius mass spin rdown rup\n",
              N, K, n + 1);

      for(j = 0; j < n + 1; ++j)
         {
            fprintf(output_file, "%le %le %le ", v[index_arr[j + 1]].r,
                    v[index_arr[j + 1]].m, v[index_arr[j + 1]].s);

            fprintf(output_file, "%le %le\n", v[index_arr[j + 1]].rdown,
                    v[index_arr[j + 1]].rup);
         }

      fflush(output_file);

      if(output_name[0] != '\n')
         {
            fclose(output_file);
         }

      return;
   }

void shift_add(long i, struct solar dust, long *n)

       /* shift present bodies and add dust body to the i-th position */
                                                         /* ver. 4.01 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
   {
      if(i <= *n)
         {
            memmove(&v[i + 1], &v[i],
                    (*n - i + 1) * sizeof(struct solar));
         }

      v[i] = dust;
      ++*n;
      return;
   }

int cond_dust(long i, struct solar dust, long *n, double K)

           /* check if dust body can merge with i-th body and further */
                                                         /* ver. 4.01 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
   {
      int dust_eat(long, struct solar*, long*, double);
      long l, left = i, right = i + 1;

      if(dust_eat(left, &dust, n, K))            /* try to merge with */
         {                                       /* the i-th body     */
            --left;                              /* success           */
            l = 1;
         }

      else
         {
            return 0;                              /* failure; return */
         }

      while(l)                           /* check for further merging */
         {
            while(dust_eat(left, &dust, n, K)) /* try to the left     */
               {                               /* as much as possible */
                  --left;
               }

            l = 0;

            while(dust_eat(right, &dust, n, K)) /* try to the right   */
               {                               /* as much as possible */
                  ++right;
                  ++l;
               }
         }

      v[left + 1] = dust;               /* update the array of bodies */

      if(right > left + 2)
         {
            memmove(&v[left + 2], &v[right],
                    (*n - right + 1) * sizeof(struct solar));

            *n -= right - left - 2;
         }

      return 1;
   }

int dust_eat(long i, struct solar *dust, long *n, double K)

                            /* trys merging dust with i-th body based */
                                      /* on the interaction criterion */
                                                          /* ver. 4.1 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
   {
      double m, r, rs, ris, f;

      if((i < 0) || (i > *n))
         {
            return 0;
         }

      if(dust -> r == v[i].r)
         {
            dust -> m = v[i].m + dust -> m;
            dust -> s += v[i].s;

            dust -> rdown = v[i].rdown < dust -> rdown ?
                            v[i].rdown : dust -> rdown;

            dust -> rup = v[i].rup > dust -> rup ?
                          v[i].rup : dust -> rup;

            return 1;
         }

      m = v[i].m + dust -> m;
      r = v[i].r < dust -> r ? v[i].r : dust -> r;
      rs = sqrt(r);
      ris = 1 / rs;

      f = (v[i].r - dust -> r) * (v[i].ris - dust -> ris) *
          (m * ris - v[i].m * v[i].ris - dust -> m * dust -> ris);

      f = f < 0 ? - f : f;

      if(f <= K * v[i].m * dust -> m)
         {
            dust -> s += v[i].s + v[i].m * v[i].rs +
                         dust -> m * dust -> rs - m * rs;

            dust -> m = m;
            dust -> r = r;
            dust -> rs = rs;
            dust -> ris = ris;

            dust -> rdown = v[i].rdown < dust -> rdown ?
                            v[i].rdown : dust -> rdown;

            dust -> rup = v[i].rup > dust -> rup ?
                          v[i].rup : dust -> rup;

            return 1;
         }

      else
         {
            return 0;
         }
   }

int save_or_not(long *isave, double Nsave)

                                       /* save to crash file or not ? */
                                                         /* ver. 1.01 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
   {
      if(++*isave == (long) Nsave)
         {
            *isave = 0;
            return 1;
         }

      else
         {
            return 0;
         }
   }

void save_crash_file(long iN, long iK, long irep, double N, double K,
                     long n, double Nsave, long seed, double i0,
                     long isave, long tsnap)

                                              /* saving to crash file */
                                                          /* ver. 1.1 */
                                                      /* Sep 28, 1999 */
                                                   /* author A. Balaz */
   {
      FILE *crash;

      system("mv crash crash.back");
      system("cp .crash crash");
      crash = fopen("crash", "ab+");
      fprintf(crash, "iN0=%ld\niK0=%ld\nirep0=%ld\n", iN, iK, irep);
      fprintf(crash, "N=%le\nK=%le\nNsave=%le\n", N, K, Nsave);

      fprintf(crash, "idum2=%ld\niy=%ld\niv[]=\n", idum2, iy);

      fwrite(iv, sizeof(long), NTAB, crash);

      fprintf(crash, "seed=%ld\ni0=%le\nisave=%ld\ntsnap=%ld\n",
              seed, i0, isave, tsnap);

      fprintf(crash, "n=%ld\nradius mass spin rdown rup (binary)\n", n);
      fwrite(v, sizeof(struct solar), n + 1, crash);
      fflush(crash);
      fclose(crash);
   }

int snap_or_not(long *tsnap, double i, int mode_snap, double Nsnapmin,
                double Nsnapstep, long Nsnaprep)

                                          /* create snapshot or not ? */
                                                          /* ver. 1.1 */
                                                      /* Sep 28, 1999 */
                                                   /* author A. Balaz */
   {
      double Nsnap;

      if(mode_snap == 0)
         {
            Nsnap = Nsnapmin;
         }

      if(mode_snap == 1)
         {
            Nsnap = Nsnapmin + *tsnap * Nsnapstep;
         }

      if(mode_snap == 2)
         {
            Nsnap = pow(10, Nsnapmin + *tsnap * Nsnapstep);
         }

      if(i < Nsnap)
         {
            return 0;
         }

      else if(i == Nsnap)
         {
            if((mode_snap == 1) || (mode_snap == 2))
               {
                  if((*tsnap < Nsnaprep) || (Nsnaprep == 0))
                     {
                        ++*tsnap;
                     }
               }

            return 1;
         }

      else
         {
            return 0;
         }
   }

void save_snap_file(double N, double K, long n, double i, long irep)

                                                   /* create snapshot */
                                                         /* ver. 1.01 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
   {
      char snap_name[SIZE];
      FILE *snap_file;
      int j;

      sprintf(snap_name, "snapi%leN%leK%ler%ld", i, N, K, irep);
      snap_file = fopen(snap_name, "w");
      fprintf(snap_file, "i=%le\nN=%le\nK=%le\nn=%ld number of final particles\n", i, N, K, n);
      fprintf(snap_file, "irep=%ld\nradius mass spin rdown rup\n", irep);

      for(j = 0; j < n + 1; ++j)
         {
            fprintf(snap_file, "%le %le %le %le %le\n", v[j].r, v[j].m,
                    v[j].s, v[j].rdown, v[j].rup);
         }

      fflush(snap_file);
      fclose(snap_file);
   }

void solar_crash(char *crash_name, char *output_name)

       /* finishes the crashed accretion, no avelanches, left merging */
                            /* for comments, see function solar above */
                                                          /* ver. 4.1 */
                                                      /* Sep 27, 1999 */
                                                   /* author A. Balaz */
  {
      double rho(long*);
      int cond_dust(long, struct solar, long*, double);
      void shift_add(long, struct solar, long*);
      void sort2(unsigned long, double*, long*);
      int save_or_not(long*, double);

      void save_crash_file(long, long, long, double, double, long,
                           double, long, double, long, long);

      int snap_or_not(long*, double, int, double, double, long);
      void save_snap_file(double, double, long, double, long);
      FILE *crash, *output_file;

      struct solar dust;
      long isave, tsnap, j;
      double i; 
      long n; 
      long down, center, up, index_arr[NMAX];
      double sort_arr[NMAX];

      int save, snap, mode_snap;             /* stuff from crash file */
      long seed, ldummy, iN, iK;
      long irep, Nsnaprep;
      double i0, ddummy, N, K, Nsave;
      double Nsnapmin, Nsnapstep;

      crash = fopen(crash_name, "rb");

      if(output_name[0] == '\n')
         {
            output_file = stdout;
         }

      else
         {
            output_file = fopen(output_name, "a+");
         }

      fscanf(crash, "mode_N=%ld\nNmin=%le\nNstep=%le\nNrep=%ld\n",
             &ldummy, &ddummy, &ddummy, &ldummy);

      fscanf(crash, "mode_K=%ld\nKmin=%le\nKstep=%le\nKrep=%ld\n",
             &ldummy, &ddummy, &ddummy, &ldummy);

      fscanf(crash, "rep=%ld\nseed=%ld\nsave=%ld\nsave_relative=%ld\n",
             &ldummy, &ldummy, &save, &ldummy);

      fscanf(crash, "Nfac=%le\nsnap=%ld\nmode_snap=%ld\n",
             &ddummy, &snap, &mode_snap);

      fscanf(crash, "Nsnapmin=%le\nNsnapstep=%le\nNsnaprep=%ld\n",
             &Nsnapmin, &Nsnapstep, &Nsnaprep);

      fscanf(crash, "iN0=%ld\niK0=%ld\nirep0=%ld\n", &iN, &iK, &irep);
      fscanf(crash, "N=%le\nK=%le\nNsave=%le\n", &N, &K, &Nsave);

      fscanf(crash, "idum2=%ld\niy=%ld\niv[]=\n", &idum2, &iy);

      fread(iv, sizeof(long), NTAB, crash);

      fscanf(crash, "seed=%ld\ni0=%le\nisave=%ld\ntsnap=%ld\n",
             &seed, &i0, &isave, &tsnap);

      fscanf(crash, "n=%ld\nradius mass spin rdown rup (binary)\n", &n);
      fread(v, sizeof(struct solar), n + 1, crash);
      fclose(crash);

      dust.m = 1 / N;
      dust.s = 0;

      for(i = i0; i < N; ++i)
         {
            if(save == 1)
               {
                  if(save_or_not(&isave, Nsave) == 1)
                     {
                        save_crash_file(iN, iK, irep, N, K, n, Nsave,
                                        seed, i, isave - 1, tsnap);
                     }
               }

            if(snap == 1)
               {
                  if(snap_or_not(&tsnap, i, mode_snap,
                                 Nsnapmin, Nsnapstep, Nsnaprep) == 1)
                     {
                        save_snap_file(N, K, n, i, irep);
                     }
               }

            dust.r = rho(&seed);
            dust.rs = sqrt(dust.r);
            dust.ris = 1 / dust.rs;
            dust.rdown = dust.r;
            dust.rup = dust.r;

            if(dust.r <= v[0].r)
               {
                  if(cond_dust(0, dust, &n, K))
                     {
                        continue;
                     }

                  else
                     {
                        shift_add(0, dust, &n);
                        continue;
                     }
               }

            if(dust.r >= v[n].r)
               {
                  if(cond_dust(n, dust, &n, K))
                     {
                        continue;
                     }

                  else
                     {
                        shift_add(n + 1, dust, &n);
                        continue;
                     }
               }

            up = n;
            down = 0;

            while(up - down > 1)
               {
                  center = (up + down) / 2;
                  
                  if(dust.r < v[center].r)
                     {
                        up = center;
                     }
                     
                  else
                     {
                        down = center;
                     }
               }

            if (cond_dust(down, dust, &n, K))
               {
                  continue;
               }

            if (cond_dust(down + 1, dust, &n, K))
               {
                  continue;
               }

            shift_add(down + 1, dust, &n);
         }

      for (j = 0; j < n + 1; ++j)
         {
            index_arr[j + 1] = j;
            sort_arr[j + 1] = v[j].m;
         }

      sort2(n + 1, sort_arr, index_arr);

      fprintf(output_file, "\nN=%le\nK=%le\nn=%ld number of final particles\nradius mass spin rdown rup\n",
              N, K, n + 1);

      for(j = 0; j < n + 1; ++j)
         {
            fprintf(output_file, "%le %le %le ", v[index_arr[j + 1]].r,
                    v[index_arr[j + 1]].m, v[index_arr[j + 1]].s);

            fprintf(output_file, "%le %le\n", v[index_arr[j + 1]].rdown,
                    v[index_arr[j + 1]].rup);
         }

      fflush(output_file);

      if(output_name[0] != '\n')
         {
            fclose(output_file);
         }

      return;
   }
